@@ -1,6 +1,6 @@
 /********************************************************************************
 * ReactPhysics3D physics library, http://www.reactphysics3d.com                 *
-* Copyright (c) 2010-2019 Daniel Chappuis                                       *
+* Copyright (c) 2010-2020 Daniel Chappuis                                       *
 *********************************************************************************
 *                                                                               *
 * This software is provided 'as-is', without any express or implied warranty.   *
@@ -27,8 +27,8 @@
 #define REACTPHYSICS3D_SET_H
 
 // Libraries
-#include "memory/MemoryAllocator.h"
-#include "mathematics/mathematics_functions.h"
+#include <reactphysics3d/memory/MemoryAllocator.h>
+#include <reactphysics3d/mathematics/mathematics_functions.h>
 #include <cstring>
 #include <stdexcept>
 #include <functional>
@@ -42,7 +42,7 @@ namespace reactphysics3d {
  * This class represents a simple generic set. This set is implemented
  * with a hash table.
   */
-template<typename V>
+template<typename V, class Hash = std::hash<V>, class KeyEqual = std::equal_to<V>>
 class Set {
 
     private:
@@ -206,11 +206,12 @@ class Set {
 
             if (mCapacity > 0) {
 
-               size_t hashCode = std::hash<V>()(value);
+               size_t hashCode = Hash()(value);
                int bucket = hashCode % mCapacity;
+                auto keyEqual  = KeyEqual();
 
                for (int i = mBuckets[bucket]; i >= 0; i = mEntries[i].next) {
-                   if (mEntries[i].hashCode == hashCode && (*mEntries[i].value) == value) {
+                   if (mEntries[i].hashCode == hashCode && keyEqual(*mEntries[i].value, value)) {
                        return i;
                    }
                }
@@ -239,32 +240,6 @@ class Set {
             return number;
         }
 
-        /// Clear and reset the set
-        void reset() {
-
-            // If elements have been allocated
-            if (mCapacity > 0) {
-
-                // Clear the list
-                clear();
-
-                // Destroy the entries
-                for (int i=0; i < mCapacity; i++) {
-                    mEntries[i].~Entry();
-                }
-
-                mAllocator.release(mBuckets, mCapacity * sizeof(int));
-                mAllocator.release(mEntries, mCapacity * sizeof(Entry));
-
-                mNbUsedEntries = 0;
-                mNbFreeEntries = 0;
-                mCapacity = 0;
-                mBuckets = nullptr;
-                mEntries = nullptr;
-                mFreeIndex = -1;
-            }
-        }
-
     public:
 
         /// Class Iterator
@@ -352,7 +327,7 @@ class Set {
                 }
 
                 /// Pre increment (++it)
-                Iterator operator++(int number) {
+                Iterator operator++(int) {
                     Iterator tmp = *this;
                     advance();
                     return tmp;
@@ -391,7 +366,7 @@ class Set {
         }
 
         /// Copy constructor
-        Set(const Set<V>& set)
+        Set(const Set<V, Hash, KeyEqual>& set)
           :mNbUsedEntries(set.mNbUsedEntries), mNbFreeEntries(set.mNbFreeEntries), mCapacity(set.mCapacity),
            mBuckets(nullptr), mEntries(nullptr), mAllocator(set.mAllocator), mFreeIndex(set.mFreeIndex) {
 
@@ -422,7 +397,7 @@ class Set {
         /// Destructor
         ~Set() {
 
-            reset();
+            clear(true);
         }
 
         /// Allocate memory for a given number of elements
@@ -445,26 +420,29 @@ class Set {
             return findEntry(value) != -1;
         }
 
-        /// Add a value into the set
-        void add(const V& value) {
+        /// Add a value into the set.
+        /// Returns true if the item has been inserted and false otherwise.
+        bool add(const V& value) {
 
             if (mCapacity == 0) {
                 initialize(0);
             }
 
             // Compute the hash code of the value
-            size_t hashCode = std::hash<V>()(value);
+            size_t hashCode = Hash()(value);
 
             // Compute the corresponding bucket index
             int bucket = hashCode % mCapacity;
 
+            auto keyEqual  = KeyEqual();
+
             // Check if the item is already in the set
             for (int i = mBuckets[bucket]; i >= 0; i = mEntries[i].next) {
 
                 // If there is already an item with the same value in the set
-                if (mEntries[i].hashCode == hashCode && (*mEntries[i].value) == value) {
+                if (mEntries[i].hashCode == hashCode && keyEqual(*mEntries[i].value, value)) {
 
-                    return;
+                    return false;
                 }
             }
 
@@ -500,6 +478,8 @@ class Set {
             assert(mEntries[entryIndex].value != nullptr);
             new (mEntries[entryIndex].value) V(value);
             mBuckets[bucket] = entryIndex;
+
+            return true;
         }
 
         /// Remove the element pointed by some iterator
@@ -517,12 +497,13 @@ class Set {
 
             if (mCapacity > 0) {
 
-                size_t hashcode = std::hash<V>()(value);
+                size_t hashcode = Hash()(value);
+                auto keyEqual  = KeyEqual();
                 int bucket = hashcode % mCapacity;
                 int last = -1;
                 for (int i = mBuckets[bucket]; i >= 0; last = i, i = mEntries[i].next) {
 
-                    if (mEntries[i].hashCode == hashcode && (*mEntries[i].value) == value) {
+                    if (mEntries[i].hashCode == hashcode && keyEqual(*mEntries[i].value, value)) {
 
                         if (last < 0 ) {
                            mBuckets[bucket] = mEntries[i].next;
@@ -561,8 +542,22 @@ class Set {
             return end();
         }
 
+        /// Return a list with all the values of the set
+        List<V> toList(MemoryAllocator& listAllocator) const {
+
+            List<V> list(listAllocator);
+
+            for (int i=0; i < mCapacity; i++) {
+                if (mEntries[i].value != nullptr) {
+                    list.add(*(mEntries[i].value));
+                }
+            }
+
+           return list;
+        }
+
         /// Clear the set
-        void clear() {
+        void clear(bool releaseMemory = false) {
 
             if (mNbUsedEntries > 0) {
 
@@ -581,6 +576,22 @@ class Set {
                 mNbFreeEntries = 0;
             }
 
+            // If elements have been allocated
+            if (releaseMemory && mCapacity > 0) {
+
+                // Destroy the entries
+                for (int i=0; i < mCapacity; i++) {
+                    mEntries[i].~Entry();
+                }
+
+                mAllocator.release(mBuckets, mCapacity * sizeof(int));
+                mAllocator.release(mEntries, mCapacity * sizeof(Entry));
+
+                mCapacity = 0;
+                mBuckets = nullptr;
+                mEntries = nullptr;
+            }
+
             assert(size() == 0);
         }
 
@@ -604,11 +615,12 @@ class Set {
 
             if (mCapacity > 0) {
 
-               size_t hashCode = std::hash<V>()(value);
+               size_t hashCode = Hash()(value);
                bucket = hashCode % mCapacity;
+               auto keyEqual  = KeyEqual();
 
                for (int i = mBuckets[bucket]; i >= 0; i = mEntries[i].next) {
-                   if (mEntries[i].hashCode == hashCode && *(mEntries[i].value) == value) {
+                   if (mEntries[i].hashCode == hashCode && keyEqual(*(mEntries[i].value), value)) {
                        entry = i;
                        break;
                    }
@@ -650,8 +662,8 @@ class Set {
             // Check for self assignment
             if (this != &set) {
 
-                // Reset the set
-                reset();
+                // Clear the set
+                clear(true);
 
                 if (set.mCapacity > 0) {
 
@@ -715,15 +727,15 @@ class Set {
         }
 };
 
-template<typename V>
-const int Set<V>::PRIMES[NB_PRIMES] = {3, 7, 11, 17, 23, 29, 37, 47, 59, 71, 89, 107, 131, 163, 197, 239, 293, 353, 431, 521, 631, 761, 919,
+template<typename V, class Hash, class KeyEqual>
+const int Set<V, Hash, KeyEqual>::PRIMES[NB_PRIMES] = {3, 7, 11, 17, 23, 29, 37, 47, 59, 71, 89, 107, 131, 163, 197, 239, 293, 353, 431, 521, 631, 761, 919,
                              1103, 1327, 1597, 1931, 2333, 2801, 3371, 4049, 4861, 5839, 7013, 8419, 10103, 12143, 14591,
                              17519, 21023, 25229, 30293, 36353, 43627, 52361, 62851, 75431, 90523, 108631, 130363, 156437,
                              187751, 225307, 270371, 324449, 389357, 467237, 560689, 672827, 807403, 968897, 1162687, 1395263,
                              1674319, 2009191, 2411033, 2893249, 3471899, 4166287, 4999559};
 
-template<typename V>
-int Set<V>::LARGEST_PRIME = -1;
+template<typename V, class Hash, class KeyEqual>
+int Set<V, Hash, KeyEqual>::LARGEST_PRIME = -1;
 
 }
 
